! Program to solve 1-D steady heat equation with preconditioned conjugate gradient
real function dot_prod(x,y,n)
integer,intent(in) :: n
real,intent(in) :: x(n),y(n)
integer :: i
dot_prod = 0.0
do i = 1,n
  dot_prod = dot_prod + x(i)*y(i)
enddo
return
end
!
subroutine residual(aw,ap,ae,u,b,res,n)
integer,intent(in) :: n
real,intent(in) :: aw(n),ap(n),ae(n),u(n),b(n)
real,intent(out) :: res(n)
integer :: i
external dot_prod
res(1) = b(1) - ap(1)*u(1) - ae(1)*u(2)
res(n) = b(n) - aw(n)*u(n-1) - ap(n)*u(n)
do i = 2,n-1
  res(i) = b(i) - aw(i)*u(i-1) - ap(i)*u(i) - ae(i)*u(i+1)
enddo
return
end
!
subroutine mat_vec(aw,ap,ae,x,ax,n)
integer,intent(in) :: n
real,intent(in) :: aw(n),ap(n),ae(n),x(n)
real,intent(out) :: ax(n)
integer :: i
ax(1) = ap(1)*x(1) + ae(1)*x(2)
ax(n) = aw(n)*x(n-1) + ap(n)*x(n)
do i = 2,n-1
  ax(i) =  aw(i)*x(i-1) + ap(i)*x(i) + ae(i)*x(i+1)
enddo
return
end
 


!page 217 Kincaid-Cheney
program heat_eqn_1D_pcg
integer :: n,nit_max
real :: delta,eps
parameter (n=10,nit_max = 100,delta=0.0000001,eps=0.0000001)
real :: aw(n),ap(n),ae(n),b(n),r(n),v(n),z(n),ax(n),av(n),x(n)
real :: c,d, ta=1.0,tb=0.0,rdotr,vdotv,vdotz,t,e
integer :: k
data aw/n*1/,ap/n*-2.0/,ae/n*1/
data b/n*0.0/,r/n*0.0/,z/n*0.0/,v/n*0.0/,x/n*0.0/
! the boundary conditions
b(1) = -ta ; b(n) = -tb
call mat_vec(aw,ap,ae,x,ax,n)
r = b - ax
z = r/ap
v = z
c = dot_prod(z,r,n)
do k = 1,nit_max
  vdotv = dot_prod(v,v,n)
  if (sqrt(vdotv) .le. delta) then
    go to 100
  else
    call mat_vec(aw,ap,ae,v,av,n)
    z = av
    vdotz = dot_prod(v,z,n)
    t = c/vdotz
    x = x + t*v
    r = r - t*z
    z = r/ap
    d = dot_prod(z,r,n)
    if (d*d .le. eps) then
      e = dot_prod(r,r,n)
      if ( e*e .le. eps) then
        go to 100
      endif
    endif
    v = z + d*v/c
    c = d
    write(*,*)'k,x,r',k,x,r
  endif
enddo
100 continue
if ( k .gt. nit_max) then
  write(*,*)'Iterations did not converge'
else
  write(*,*)'Converged in',k,' iterations'
endif
stop
end
